# XXX: Remove this - we want to push this down into Zope

# For ZPublisher.interfaces

from zope.interface import Interface, Attribute

class IPubBeforeStreaming(Interface):
    """Event fired just before a streaming response is initiated, i.e. when
    something calls response.write() for the first time. Note that this is
    carries a reference to the *response*, not the request.
    """
    
    response = Attribute(u"The current HTTP response")

import ZPublisher.interfaces
setattr(ZPublisher.interfaces, 'IPubBeforeStreaming', IPubBeforeStreaming)

# For ZPublisher.pubevents

from zope.interface import implements

class PubBeforeStreaming(object):
    """Notified immediately before streaming via response.write() commences
    """
    implements(IPubBeforeStreaming)
    
    def __init__(self, response):
        self.response = response

import ZPublisher.pubevents
setattr(ZPublisher.pubevents, 'PubBeforeStreaming', PubBeforeStreaming)

# Patch for ZPublisher.HTTPResponse

from zope.event import notify

def ZPublisher_HTTPResponse_write(self, data):
    if not self._wrote:
        
        notify(PubBeforeStreaming(self))
        
        self.outputBody()
        self._wrote = 1
        self.stdout.flush()

    self.stdout.write(data)

# Patch for ZServer.HTTPResponse

# from zope.event import notify
import tempfile
import thread
from ZServer.Producers import file_part_producer

def ZServer_ZServerHTTPResponse_write(self,data):
    if type(data) != type(''):
        raise TypeError('Value must be a string')

    stdout=self.stdout

    if not self._wrote:
        
        notify(PubBeforeStreaming(self))
        
        l=self.headers.get('content-length', None)
        if l is not None:
            try:
                if type(l) is type(''): l=int(l)
                if l > 128000:
                    self._tempfile=tempfile.TemporaryFile()
                    self._templock=thread.allocate_lock()
            except: pass

        self._streaming=1
        stdout.write(str(self))
        self._wrote=1

    if not data: return

    if self._chunking:
        data = '%x\r\n%s\r\n' % (len(data),data)

    l=len(data)

    t=self._tempfile
    if t is None or l<200:
        stdout.write(data)
    else:
        b=self._tempstart
        e=b+l
        self._templock.acquire()
        try:
            t.seek(b)
            t.write(data)
        finally:
            self._templock.release()
        self._tempstart=e
        stdout.write(file_part_producer(t,self._templock,b,e), l)



import unittest

import zope.component.testing

from zope.interface import implements
from zope.interface import Interface
from z3c.caching.registry import RulesetRegistry

from zope.component import provideUtility
from zope.component import provideAdapter
from zope.component import adapts

from plone.registry.interfaces import IRegistry
from plone.registry import Registry, Record
from plone.registry import field

from plone.caching.operations import Chain
from plone.caching.interfaces import ICachingOperation

_marker = object()

class DummyView(object):
    pass

class DummyResponse(dict):
    
    def addHeader(self, name, value):
        self.setdefault(name, []).append(value)
    
    def setHeader(self, name, value):
        self[name] = value

class DummyRequest(dict):
    def __init__(self, published, response):
        self['PUBLISHED'] = published
        self.response = response


class TestChain(unittest.TestCase):
    
    def setUp(self):
        self.registry = Registry()
        provideUtility(self.registry, IRegistry)
        provideAdapter(RulesetRegistry)
    
    def tearDown(self):
        zope.component.testing.tearDown()
    
    def test_no_option(self):
        view = DummyView()
        request = DummyRequest(view, DummyResponse())
        
        chain = Chain(view, request)
        ret = chain.interceptResponse('testrule', request.response)
        chain.modifyResponse('testrule', request.response)
        
        self.assertEquals(None, ret)
        self.assertEquals({'PUBLISHED': view}, dict(request))
        self.assertEquals({}, dict(request.response))
    
    def test_operations_list_not_set(self):
        
        self.registry.records["%s.operations" % Chain.prefix] = Record(field.List())
        
        view = DummyView()
        request = DummyRequest(view, DummyResponse())
        
        chain = Chain(view, request)
        ret = chain.interceptResponse('testrule', request.response)
        chain.modifyResponse('testrule', request.response)
        
        self.assertEquals(None, ret)
        self.assertEquals({'PUBLISHED': view}, dict(request))
        self.assertEquals({}, dict(request.response))
    
    def test_operations_empty(self):
        
        self.registry.records["%s.operations" % Chain.prefix] = Record(field.List(), [])
        
        view = DummyView()
        request = DummyRequest(view, DummyResponse())
        
        chain = Chain(view, request)
        ret = chain.interceptResponse('testrule', request.response)
        chain.modifyResponse('testrule', request.response)
        
        self.assertEquals(None, ret)
        self.assertEquals({'PUBLISHED': view}, dict(request))
        self.assertEquals({}, dict(request.response))
    
    def test_chained_operations_not_found(self):
        
        self.registry.records["%s.operations" % Chain.prefix] = Record(field.List(), ['op1'])
        
        view = DummyView()
        request = DummyRequest(view, DummyResponse())
        
        chain = Chain(view, request)
        chain.modifyResponse('testrule', request.response)
        
        self.assertEquals({'PUBLISHED': view}, dict(request))
        self.assertEquals({}, dict(request.response))
    
    def test_multiple_operations_one_found(self):
        self.registry.records["%s.operations" % Chain.prefix] = Record(field.List(), ['op1', 'op2'])
        
        view = DummyView()
        request = DummyRequest(view, DummyResponse())
        
        class DummyOperation(object):
            implements(ICachingOperation)
            adapts(Interface, Interface)
            
            def __init__(self, published, request):
                self.published = published
                self.request = request
            
            def interceptResponse(self, rulename, response):
                return u"foo"
            
            def modifyResponse(self, rulename, response):
                response['X-Mutated'] = rulename
        
        provideAdapter(DummyOperation, name="op2")
        
        chain = Chain(view, request)
        ret = chain.interceptResponse('testrule', request.response)
        
        self.assertEquals(u"foo", ret)
        self.assertEquals({'PUBLISHED': view}, dict(request))
        self.assertEquals({'X-Cache-Chain-Operations': 'op2'}, dict(request.response))
        
        request = DummyRequest(view, DummyResponse())
        chain = Chain(view, request)
        chain.modifyResponse('testrule', request.response)
        
        self.assertEquals({'PUBLISHED': view}, dict(request))
        self.assertEquals({'X-Mutated': 'testrule',
                           'X-Cache-Chain-Operations': 'op2'}, dict(request.response))

    def test_multiple_operations_multiple_found(self):
        self.registry.records["%s.operations" % Chain.prefix] = Record(field.List(), ['op1', 'op2'])
        
        view = DummyView()
        request = DummyRequest(view, DummyResponse())
        
        class DummyOperation1(object):
            implements(ICachingOperation)
            adapts(Interface, Interface)
            
            def __init__(self, published, request):
                self.published = published
                self.request = request
            
            def interceptResponse(self, rulename, response):
                return u"foo"
            
            def modifyResponse(self, rulename, response):
                response['X-Mutated-1'] = rulename
        
        provideAdapter(DummyOperation1, name="op1")
        
        class DummyOperation2(object):
            implements(ICachingOperation)
            adapts(Interface, Interface)
            
            def __init__(self, published, request):
                self.published = published
                self.request = request
            
            def interceptResponse(self, rulename, response):
                return u"bar"
            
            def modifyResponse(self, rulename, response):
                response['X-Mutated-2'] = rulename
        
        provideAdapter(DummyOperation2, name="op2")
        
        chain = Chain(view, request)
        ret = chain.interceptResponse('testrule', request.response)
        
        self.assertEquals(u"foo", ret)
        self.assertEquals({'PUBLISHED': view}, dict(request))
        self.assertEquals({'X-Cache-Chain-Operations': 'op1'}, dict(request.response))
        
        request = DummyRequest(view, DummyResponse())
        chain = Chain(view, request)
        chain.modifyResponse('testrule', request.response)
        
        self.assertEquals({'PUBLISHED': view}, dict(request))
        self.assertEquals({'X-Mutated-1': 'testrule',
                           'X-Mutated-2': 'testrule',
                           'X-Cache-Chain-Operations': 'op1; op2'}, dict(request.response))

def test_suite():
    return unittest.defaultTestLoader.loadTestsFromName(__name__)



from zope.interface import implements, classProvides, Interface
from zope.component import adapts, queryMultiAdapter

from plone.caching.interfaces import ICachingOperation
from plone.caching.interfaces import ICachingOperationType
from plone.caching.interfaces import _

from plone.caching.utils import lookupOptions

class Chain(object):
    """Caching operation which chains together several other operations.
    
    When intercepting the response, the first chained operation to return a
    value will be used. Subsequent operations are ignored. When modifying the
    response, all operations will be called, in order.
    
    The names of the operations to execute are found in the ``plone.registry``
    option ``plone.caching.operations.chain.operations`` by default, and can
    be customised on a per-rule basis with
    ``plone.caching.operations.chain.${rulename}.chain``.
    
    The option must be a sequence type (e.g. a ``Tuple``).
    """
    
    implements(ICachingOperation)
    adapts(Interface, Interface)
    
    # Type metadata
    classProvides(ICachingOperationType)
    
    title = _(u"Chain")
    description = _(u"Allows multiple operations to be chained together")
    prefix = 'plone.caching.operations.chain'
    options = ('operations',)
    
    def __init__(self, published, request):
        self.published = published
        self.request = request

    def interceptResponse(self, rulename, response):
        options = lookupOptions(self.__class__, rulename)
        
        chained = []
        
        if options['operations']:
            for name in options['operations']:
                
                operation = queryMultiAdapter((self.published, self.request),
                                             ICachingOperation, name=name)
                
                if operation is not None:
                    chained.append(name)
                    
                    value = operation.interceptResponse(rulename, response)
                    if value is not None:
                        response.setHeader('X-Cache-Chain-Operations', '; '.join(chained))
                        return value
    
    def modifyResponse(self, rulename, response):
        options = lookupOptions(self.__class__, rulename)
        
        chained = []
        
        if options['operations']:
            for name in options['operations']:
                
                operation = queryMultiAdapter((self.published, self.request),
                                             ICachingOperation, name=name)
                
                if operation is not None:
                    chained.append(name)
                    operation.modifyResponse(rulename, response)

        if chained:
            response.setHeader('X-Cache-Chain-Operations', '; '.join(chained))
